#!python
# Gap sequences inferred by IgDiscover

# Copyright (c) 2021 William Lees

# This source code, and any executable file compiled or derived from it, is governed by the European Union Public License v. 1.2,
# the English version of which is available here: https://perma.cc/DK5U-NDVE


from receptor_utils import number_v
import argparse
from receptor_utils import simple_bio_seq as simple
import Bio.Data.CodonTable


def get_parser():
    parser = argparse.ArgumentParser(description='Gap the sequences in a FASTA file, using the closest sequence in a gapped reference set as template')
    parser.add_argument('inferred_file', help='ungapped inferred sequences (fasta)')
    parser.add_argument('ref_file', help='gapped reference set (fasta)')
    parser.add_argument('out_file', help='output file containing gapped inferred sequences')
    return parser


def main():
    args = get_parser().parse_args()

    inferred = simple.read_fasta(args.inferred_file)
    refs = simple.read_fasta(args.ref_file)
    
    # Check that reference sequences have conserved residues - remove any that don't
    
    for ref, seq in list(refs.items()):
        try:
            # if the sequence has a leading gap, make it end on a codon boundary for the purpose of the checks

            trial_seq = seq.replace('.', '-')

            if trial_seq[0] == '-':
                aa = simple.translate(trial_seq, ignore_partial_codon=True)
            else:
                aa = simple.translate(trial_seq, ignore_partial_codon=False)

            notes = number_v.check_conserved_residues(aa)

            if notes:
                if 'truncated' in notes:
                    print(f"Warning: reference sequence {ref}: {notes} (if you don't want to use this as a reference sequence, please remove it from the file)")
                else:
                    print(f'Removing {ref} from reference: {notes}')
                    del refs[ref]
        except Bio.Data.CodonTable.TranslationError as e:
            print(f'Removing {ref} from reference: {e}\n{aa}')
            del refs[ref]

    ungapped_refs = {}

    for id, seq in refs.items():
        ungapped_refs[id] = seq.replace('.', '')

    gapped = {}

    for id, seq in inferred.items():
        gapped[id], aa, notes = number_v.gap_sequence(seq, refs, ungapped_refs)
        if notes:
            print('%s: %s' % (id, notes))

    simple.write_fasta(args.out_file, gapped)



if __name__ == "__main__":
    main()



"""Wrapper around BioSeq functions for simple applications
principles:
- store sequences as strings, use dicts for collectiopns
- all sequences are coerced to upper case on input
- iterators are coreced into lists for ease of debugging
basically just make things simple for cases where we don't need to do more
"""

# Copyright (c) 2021 William Lees
__author__ = 'William Lees'

# This source code, and any executable file compiled or derived from it, is governed by the European Union Public License v. 1.2,
# the English version of which is available here: https://perma.cc/DK5U-NDVE

from warnings import warn
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Align import MultipleSeqAlignment, AlignInfo
import random
import csv
from collections import Counter


def chunks(l: str, n: int):
    """ Yield successive n-sized chunks from l.
    """
    for i in range(0, len(l), n):
        yield l[i:i+n]


def read_fasta(infile: str):
    """Read a FASTA file into a dict

    :param infile: Pathname of the file
    :type infile: str
    :returns: A dictionary indexed by FASTA ID, containing the sequences (in upper case)
    :rtype: dict
    """
    res = {}
    recs = SeqIO.parse(infile, 'fasta')
    for rec in recs:
        res[rec.id] = str(rec.seq).upper()
    return res


def write_fasta(outfile: str, seqs: dict):
    """Write a dict into a FASTA file. The dict should be indexed by sequence name

    :param outfile: Pathname of the file to be written
    :type outfile: str
    :param seqs: The sequences to write
    :type seqs: dict
    :returns: the number of records written
    :rtype: int
    """
    try:
        SeqIO.write(toSeqRecords(seqs), outfile, 'fasta')
    except:
        SeqIO.write(toSeqRecords(outfile), seqs, 'fasta')
        warn("write_fasta(seqs, outfile) is deprecated. Please use  write_fasta(outfile, seqs) instead", DeprecationWarning, stacklevel=2)


def read_single_fasta(infile):
    """Read a single sequence from a FASTA file

    :param infile: Pathname of the file
    :type infile: str
    :returns: The first (or only) sequence in the file
    :rtype: str
    """
    recs = SeqIO.parse(infile, 'fasta')
    for rec in recs:
        return str(rec.seq).upper()


# read V,D,J regions from one or more species from an IMGT reference file
# species should be a ***list*** of species
# returns a nested dict: [species][chain]
def read_imgt_fasta(infile: str, species: str, chains=('IGHV', 'IGHD', 'IGHJ', 'CH'), functional_only: bool =False, include_orphon: bool =False):
    """read V,D,J regions from one or more species from an IMGT reference file

    :param infile: The IMGT reference file
    :type infile: str
    :param species: Species as specified in the file, with spaces replaced by underscore
    :type species: str
    :param chains: List of chains to read (e.g. ['IGHV', 'IGHJ'])
    :type chains: list
    :param functional_only: If True, returns only sequences marked as functional (F)
    :param functional_only: If True, includes orphons
    :return: A dict containing the serquences, indexed by name
    :rtype: dict
    """
    res = {}
    for sp in species:
        res[sp] = {}
        for chain in chains:
            res[sp][chain] = {}

    recs = SeqIO.parse(infile, 'fasta')

    for rec in recs:
        for sp in species:
            for chain in chains:
                if ('-REGION' in rec.description or chain == 'CH' or 'EX' in rec.description) and sp in rec.description and chain in rec.description:
                    if not functional_only or '|F|' in rec.description:
                        if 'V' not in chain or len(rec.seq) > 280:       # strip out obviously incomplete Vs
                            name = rec.description.split('|')[1]
                            if include_orphon or '/OR' not in name:
                                seq_type = rec.description.split('|')[4]

                                if chain == 'CH' or ('EX' in rec.description and chain in rec.description):
                                    gene, allele = name.split('*')
                                    res[sp][chain][gene + '_' + seq_type + '*' + allele] = str(rec.seq).upper()
                                else:
                                    res[sp][chain][name] = str(rec.seq).upper()
    return res


def sample_fasta(seqs: dict, number: int):
    """Return a random sample of sequences stored in a dict. Sequences are not resampled.
    Returns a ValueError if the dict does not contain enough sequences

    :param seqs: The sequences to sample
    :type seqs: dict
    :param number: The number of sequences to return
    :type number: int
    :return: The sampled sequences
    :rtype: dict
    """
    keys = seqs.keys()

    if len(keys) <= number:
        return seqs

    sample = random.sample(keys, number)
    sample_seqs = {}
    for key in sample:
        sample_seqs[key] = seqs[key]
    return sample_seqs


def translate(seq: str, truncate: bool = True, ignore_partial_codon: bool = True):
    """Translate a nucleotide sequence to amino acid

    :param seq: the sequence to translate
    :type seq: str
    :param truncate: If True, truncate the sequence so that it terminates on a codon boundary. Otherwise pad with N if necessary
    :type truncate: bool
    :param ignore_partial_codon: If True, if any position in a codon contains - or ., set the entire codon to --- ensuring it gets translated as -
    :type ignore_partial_codon: bool
    :return: Amino acid string
    :rtype: str
    """
    seq_len = len(seq)
    residue = seq_len % 3

    if residue != 0:
        if truncate:
            seq = seq[:seq_len - residue]
        else:
            seq += 'N' * (3 - residue)
            
    if ignore_partial_codon:
        codons = [codon for codon in chunks(seq, 3)]
        i = 0

        for i in range(len(codons)):
            if '-' in codons[i] or '.' in codons[i]:
                codons[i] = '---'
            
        seq = ''.join(codons)

    return str(Seq(seq).translate())


def reverse_complement(seq: str):
    """Return the reverse complement of a nucelotide sequence

    :param seq: the nucleotide sequence
    :type seq: str
    :return: the reverse complement
    :rtype: str
    """
    return str(Seq(seq).reverse_complement())


def nt_diff(s1: str, s2: str):
    """Returns a count of the positions in which the input sequences differ. If the sequences are of different lengths, the count only covers
    positions in the shorter sequence.

    :param s1: the first sequence
    :type s1: str
    :param s2: the second sequence
    :type s2: str
    :return: The number of positions at which the sequences differ
    :rtype: int
    """
    diffs = 0
    row = list(zip(s1.upper(), s2.upper()))

    for p in range(len(row)):
        if row[p][0] != row[p][1] and (row[p][0] not in ['X', 'N'] and row[p][1] not in ['X', 'N']):
            diffs += 1

    diffs += abs(len(s1) - len(s2))
    return diffs


# convert our dict format to SeqRecords
def toSeqRecords(seqs: dict):
    """Convert a dict of sequences to a list of BioPython SeqRecords

    :param seqs: the sequences to convert
    :type seqs: dict
    :return: the BioPython SeqRecords
    :rtype: list
    """
    recs = []
    for name, seq in seqs.items():
        recs.append(SeqRecord(Seq(seq), id=name, description=''))
    return recs


def dumb_consensus(seqs: dict, threshold: float = 0.7):
    """Return a dumb consensus on a dict of sequences, using Bio.Align.AlignInfo.SummaryInfo.dumb_consensus. All sequences should be the same length.

    :param seqs: the sequences to analyse
    :type seqs: dict
    :param threshold:  the threshold value that is required to add a particular atom
    :type threshold: float
    :return: the consensus sequence
    :rtype: str
    """
    msa = MultipleSeqAlignment(toSeqRecords(seqs))
    summary_align = AlignInfo.SummaryInfo(msa)
    return str(summary_align.gap_consensus(threshold=threshold))


def scored_consensus(seqs: dict, threshold: float = 0.7):
    """Return a dumb consensus on a dict of sequences, using Bio.Align.AlignInfo.SummaryInfo.gap_consensus. All sequences should be the same length.
    The output is modified to provide, as well as the consensus string, the mimimum score achieved at any position

    :param seqs: the sequences to analyse
    :type seqs: dict
    :param threshold:  the threshold value that is required to add a particular atom
    :type threshold: float
    :return: (consensus sequence, min_score)
    :rtype: str
    """
    msa = MultipleSeqAlignment(toSeqRecords(seqs))
    summary_align = AlignInfo.SummaryInfo(msa)
    return scored_gap_consensus(msa, threshold=threshold)


# Adapted from BioPython.Bio.ALign
# return min_score: this is the lowest proportion of reads that 'won' at any position
def scored_gap_consensus(alignment, threshold=0.7, ambiguous="X", require_multiple=False):
    """Adapted from BioPython.Bio.ALign. This function is called by receptor_utils.scored_consensus.
    """
    consensus = ""

    # find the length of the consensus we are creating
    con_len = alignment.get_alignment_length()
    con_rows = len(alignment)

    # go through each seq item
    min_score = 1.0
    for n in range(con_len):
        # keep track of the counts of the different atoms we get
        atom_dict = Counter()
        num_atoms = 0

        for record in alignment:
            # make sure we haven't run past the end of any sequences
            # if they are of different lengths
            try:
                c = record[n]
            except IndexError:
                continue
            atom_dict[c] += 1

            num_atoms += 1

        max_atoms = []
        max_size = 0

        for atom in atom_dict:
            if atom_dict[atom] > max_size:
                max_atoms = [atom]
                max_size = atom_dict[atom]
            elif atom_dict[atom] == max_size:
                max_atoms.append(atom)

        if require_multiple and num_atoms == 1:
            consensus += ambiguous
            min_score = min(min_score, 1/con_rows)
        elif (len(max_atoms) == 1) and (
            (float(max_size) / float(num_atoms)) >= threshold
        ):
            consensus += max_atoms[0]
            min_score = min(min_score, float(max_size) / float(con_rows))
        else:
            consensus += ambiguous
            min_score = min(min_score, float(max_size) / float(con_rows))

    return (Seq(consensus), min_score)


# Find the closest reference sequence to a query sequence
# Returns a list of equally close reference names
def closest_ref(seq: str, ref: dict):
    """Given an input sequence, find the closest entry or entries in an IMGT or other reference set as determined by nt_diff.

    :param seq: the input sequence
    :type seq: str
    :param ref: the reference set
    :type ref: dict
    :return: the name of the closest entry, entries, if >1 sequences in the reference set were equally close
    :rtype: list
    """
    closest_diff = 99999
    closest_names = []
    for ref_name, ref_seq in ref.items():
        diff = nt_diff(seq, ref_seq)
        if diff < closest_diff:
            closest_diff = diff
            closest_names = [ref_name]
        elif diff == closest_diff:
            closest_names.append(ref_name)
    return closest_names


# Read csv file into a list of dicts
def read_csv(file: str, delimiter: str = None):
    """Read a delimited file into a list of dicts (as produced by DictReader)

    :param file: filename of the file
    :type file: str
    :param delimiter: the delimiter (',' by default)
    :type delimiter: str
    :return: the list of dicts
    :rtype: list
    """
    ret = []
    with open(file, 'r') as fi:
        if delimiter:
            reader = csv.DictReader(fi, delimiter=delimiter)
        else:
            reader = csv.DictReader(fi)
        for row in reader:
            ret.append(row)

    return ret


# Write csv file given a list of dicts. Fieldnames are taken from the first row
def write_csv(file: str, rows: list, delimiter: str = None, scan_all: bool = False):
    """Write a list of dicts to a delimited file. The header row is determined from the keys of the first item

    :param file: filename of the delimited file to create
    :type file: str
    :param rows: the rows to write
    :type rows: list
    :param delimiter: the delimiter (',' by default)
    :type delimiter: str
    :param scan_all: If True, scan all rows to find the full set of keys
    :type scan_all: bool
    :return: None
    :rtype: None
    """
    if not rows:
        return

    if scan_all:
        fieldnames = list()
        for row in rows:
            for key in row.keys():
                if key not in fieldnames:
                    fieldnames.append(key)       # use lists rather set so we can retain ordering 
    else:
        fieldnames = rows[0].keys()

    with open(file, 'w', newline='') as fo:
        if delimiter:
            writer = csv.DictWriter(fo, fieldnames=list(fieldnames), delimiter=delimiter)
        else:
            writer = csv.DictWriter(fo, fieldnames=fieldnames)

        writer.writeheader()
        
        for row in rows:
            writer.writerow(row)



#!python
# List identical sequences and sub-sequences in a fasta file

# Copyright (c) 2021 William Lees

# This source code, and any executable file compiled or derived from it, is governed by the European Union Public License v. 1.2,
# the English version of which is available here: https://perma.cc/DK5U-NDVE


import argparse
from receptor_utils import simple_bio_seq as simple


def get_parser():
    parser = argparse.ArgumentParser(description='List identical sequences and sub-sequences in a fasta file')
    parser.add_argument('input_file', help='gapped imgt reference file')
    return parser


def main():
    args = get_parser().parse_args()

    seqs = simple.read_fasta(args.input_file)

    seq_index = {}

    for seq_name, seq in seqs.items():
        if seq not in seq_index:
            seq_index[seq] = []
        seq_index[seq].append(seq_name)

    print('Identical sequences:')

    for names in seq_index.values():
        if len(names) > 1:
            print(', '.join(names))

    print('Sub-sequences:')
    for n1, s1 in seqs.items():
        for n2, s2 in seqs.items():
            if s1 != s2  and s1 in s2:
                print(f'{n1} ({s1})is a sub-sequence of {n2} ({s2})')



if __name__ == "__main__":
    main()



import argparse
import requests
import json
from Bio import SeqIO
from io import StringIO
from receptor_utils import simple_bio_seq as simple


def do_request(url):
    """
    Makes a GET request to the specified URL.

    Args:
        url (str): The URL to request.

    Returns:
        requests.Response: The response object.
    """

    # handle errors
    try:
        response = requests.get(url)
        response.raise_for_status()
    except requests.exceptions.HTTPError as err:
        print(f"HTTP error occurred: {err} {response.text}")
        quit()
    except requests.exceptions.RequestException as err:
        print(f"Request error occurred: {err}")
        quit()

    return response


def download_germline_set(species, locus, version='latest', germline_set_name=None, file_format='AIRRC-JSON', url='https://ogrdb.airr-community.org/api_v2', prefix=None):
    """
    Downloads the specified germline set using the OGRDB API.

    Args:
        species (str): The species name (e.g., "Human").
        locus (str): The locus (e.g., IGH, IGK, IGL, TRA, TRB, TRD).
        version (str): The specific version to download, default is 'latest'.
        germline_set_name (str): The name of the germline set, default is None.
        file_format (str): The format to download (AIRRC-JSON, SINGLE-FG, SINGLE-FU).
        url (str): The base URL for the API, default is 'https://ogrdb.airr-community.org/api_v2'.
        prefix (str): The prefix for filenames, default is None.
    """
    species_id = find_species_id(url, species)
    if species_id:
        germline_set_id = find_available_sets(url, species_id, locus, germline_set_name)
        if germline_set_id is not None:
            if prefix is None:
                prefix = species.replace(' ', '_') + '_' + locus + '_'
            elif prefix == 'NONE':
                prefix = ''

            if 'MULTI' in file_format:
                data_u = request_set(url, germline_set_id, version, species, germline_set_name, locus, 'SINGLE-FU')
                data_g = request_set(url, germline_set_id, version, species, germline_set_name, locus, 'SINGLE-FG')

                if not data_u or not data_g:
                    return
                
                data_u = SeqIO.parse(StringIO(data_u.text), 'fasta')
                data_u = {rec.id: str(rec.seq).upper() for rec in data_u}
                data_g = SeqIO.parse(StringIO(data_g.text), 'fasta')
                data_g = {rec.id: str(rec.seq).upper() for rec in data_g}
                files_written = []
                for chain in ['V', 'D', 'J']:
                    seqs = {k: v for k, v in data_u.items() if k[3] == chain}
                    if seqs: 
                        simple.write_fasta(f"{prefix}{chain}.fasta", seqs)
                        files_written.append(f"{prefix}{chain}.fasta")
                seqs = {k: v for k, v in data_g.items() if k[3] == 'V'}
                if seqs:
                    simple.write_fasta(f"{prefix}V_gapped.fasta", seqs)
                    files_written.append(f"{prefix}V_gapped.fasta")
                print(f"FASTA files saved to {', '.join(files_written)}")
            else:
                data = request_set(url, germline_set_id, version, species, germline_set_name, locus, file_format)
                if not data:
                    return
                if file_format == 'AIRRC-JSON':
                    output_file = f"{prefix}.json"
                    write_json_to_disk(data, output_file)
                elif file_format == 'SINGLE-FG':
                    output_file = f"{prefix}gapped.fasta"
                    write_fasta_to_disk(data.text, output_file)
                elif file_format == 'SINGLE-FU':
                    output_file = f"{prefix}gapped.fasta"
                    write_fasta_to_disk(data.text, output_file)

                


def write_fasta_to_disk(data, output_file):
    """
    Writes FASTA data to a file.

    Args:
        data (str): The FASTA data to write.
        output_file (str): The name of the output file.
    """
    with open(output_file, 'w') as f:
        f.write(data)
    
    print(f"FASTA file saved to {output_file}")


def write_json_to_disk(data, output_file):
    """
    Writes JSON data to a file.

    Args:
        data (dict): The JSON data to write.
        output_file (str): The name of the output file.
    """
    with open(output_file, 'w') as f:
        json.dump(data, f, indent=2)
    print(f"JSON file saved to {output_file}")


def request_set(url, germline_set_id, version, species, germline_set_name, locus, format):
    """
    Requests the germline set data from the API.

    Args:
        url (str): The base URL for the API.
        germline_set_id (str): The ID of the germline set.
        version (str): The version to download.
        species (str): The species name.
        germline_set_name (str): The name of the germline set.
        file_format (str): The format to download (AIRRC-JSON, SINGLE-FG, SINGLE-FU).

    Returns:
        dict or requests.Response: The requested data in JSON or FASTA format.
    """

    if version != 'latest':
        versions_list = get_set_versions(url, germline_set_id)
    else:
        versions_list = []

    if version == "latest" or version in versions_list:
        if format == 'AIRRC-JSON':
            api_url = f"{url}/germline/set/{germline_set_id}/{version}"
            response = do_request(api_url)
            return response.json()
        else:
            if format == "SINGLE-FG":
                api_url = f"{url}/germline/set/{germline_set_id}/{version}/gapped"
            elif format == "SINGLE-FU":
                api_url = f"{url}/germline/set/{germline_set_id}/{version}/ungapped"

            response = do_request(api_url)
            return response 
    else:
        gs = germline_set_name if germline_set_name else ""
        loc = locus if locus else ""
        print(f"""Version {version} of set {species} {gs} {loc} is not available. Available versions:
{', '.join(versions_list)}""")
        return None


def get_set_versions(url, germline_set_id):
    """
    Gets the available versions of a germline set.

    Args:
        url (str): The base URL for the API.
        germline_set_id (str): The ID of the germline set.

    Returns:
        list: A list of available versions.
    """
    api_url = f"{url}/germline/set/{germline_set_id}/versions"
    response = do_request(api_url)
    versions_list = [str(x) for x in response.json()['versions']]
    return versions_list


def find_available_sets(url, species_id, locus, germline_set_name):
    """
    Finds available germline sets for a species and locus.

    Args:
        url (str): The base URL for the API.
        species_id (str): The species ID.
        locus (str): The locus (e.g., IGH, IGK, IGL, TRA, TRB, TRD).
        germline_set_name (str): The name of the germline set.

    Returns:
        str: The ID of the found germline set, or None if not found.
    """
    api_url = f"{url}/germline/sets/{species_id}"
    response = do_request(api_url)
    sets_list = response.json().get('germline_species')
    found = False
    found_count = 0
    available_sets = []
    available_versions = []
    germline_set_id = None
    for set in sets_list:
        if set.get('locus') == locus:
            found = True
            found_count += 1
            available_sets.append(set['germline_set_name'])
            available_versions.append(set['germline_set_id'])
            germline_set_id = set['germline_set_id']
    
    if not found:
        available_locus = []
        for obj in sets_list:
            if obj.get('locus') not in available_locus:
                available_locus.append(obj.get('locus'))  

        print(f'''No sets for locus {locus} in species {species_id}. Available loci:
{', '.join(available_locus)}''')

    elif found_count > 1 and germline_set_name is None:
        if len(set(available_sets)) >= 1:
            germline_set_id = None
            print(f'''Multiple sets are avaliable for species {species_id} locus {locus}. 
Please specify the set name. Available set names:
{', '.join(available_sets)}''')
        else:
            return 'latest'
    else:
        print(f'{germline_set_id}')

    return germline_set_id


def find_species_id(url, species):
    """
    Finds if a species is available in the API.

    Args:
        url (str): The base URL for the API.
        species (str): The species name.

    Returns:
        bool: True if the species is found, False otherwise.
    """
    api_url = f"{url}/germline/species"
    print(api_url)
    response = do_request(api_url)

    species_list = response.json().get('species')
    id = None
    for object in species_list:
        if object.get('label') == species:
            id = object.get('id')

    if not id:
        available_species = ', '.join(obj.get('label') for obj in species_list)
        print(f'''Error: species not found. Available species:
{available_species}''')

    else:
        print(f'{species}: {id}')

    return id


def get_parser():
    parser = argparse.ArgumentParser(description='Download germline sets from the Open Germline Receptor Database (OGRDB)')
    parser.add_argument('species', type=str, help='Species (e.g. "Human")')
    parser.add_argument('locus', type=str, help='Locus (IGH, IGK, IGL, TRA, TRB, TRD, TRD)')
    parser.add_argument('-n', '--name', type=str, help='germline set name (the utility will attempt to determine the name, if none is specified)')
    parser.add_argument('-v', '--version', type=str, default='latest', help='Specific version to download, otherwise the latest version will be downloaded')
    parser.add_argument('-f', '--format', type=str, default='AIRRC-JSON', choices=['AIRRC-JSON', 'SINGLE-FG', 'SINGLE-FU', 'MULTI-F'], help='Format to download')
    parser.add_argument('-u', '--url', type=str, default='https://ogrdb.airr-community.org/api_v2', help='URL to use')
    parser.add_argument('-p', '--prefix', type=str, help='Prefix for filenames. Default prefix is species_locus (with _ substituted for space). If PREFIX is NONE, no prefix will be used for multi files, and the default prefix will be used for single files.')
    return parser


def main():
    args = get_parser().parse_args()
    download_germline_set(args.species, args.locus, args.version, args.name, args.format, args.url, args.prefix)


if __name__ == "__main__":
    main()


#!python
# Extract the sequence at the specified 1-based co-ordinates in the reference sequence

# Copyright (c) 2021 William Lees

# This source code, and any executable file compiled or derived from it, is governed by the European Union Public License v. 1.2,
# the English version of which is available here: https://perma.cc/DK5U-NDVE


import argparse
from receptor_utils import simple_bio_seq as simple


def get_parser():
    parser = argparse.ArgumentParser(description=' Extract the sequence at the specified 1-based co-ordinates in the file')
    parser.add_argument('input_file', help='File containing the reference sequence (FASTA)')
    parser.add_argument('start', help='Start coord (1-based)', type=int)
    parser.add_argument('end', help='End coord (1-based)', type=int)
    parser.add_argument('-n', help='Label of sequence in file (must specify if there is more than one sequence in the file)')
    parser.add_argument('-r', help='Reverse-complement the result', action='store_true')
    parser.add_argument('-xr', help='Extract the result from the reverse-complement of the sequence (coords refer to the reverse complement sequence)', action='store_true')
    return parser


def main():
    args = get_parser().parse_args()

    if args.end < args.start or args.end < 1 or args.start < 1:
        quit()

    if not args.n:
        ref = simple.read_single_fasta(args.input_file)
    else:
        refs = simple.read_fasta(args.input_file)
        if args.n not in refs:
            print(f'Sequence {args.n} not found in file')
            print(f'Sequences in file: {", ".join(refs.keys())}')
            quit()
        ref = refs[args.n]

    if args.xr:
        ref = simple.reverse_complement(ref)

    seq = ref[args.start-1:args.end]

    if args.r:
        seq = simple.reverse_complement(seq)
        
    print(seq)







#!python
# Extract reference files for nominated species
# The current IMGT reference file can be downloaded from https://www.imgt.org/download/GENE-DB/IMGTGENEDB-ReferenceSequences.fasta-nt-WithGaps-F+ORF+inframeP

# Copyright (c) 2021 William Lees

# This source code, and any executable file compiled or derived from it, is governed by the European Union Public License v. 1.2,
# the English version of which is available here: https://perma.cc/DK5U-NDVE


import argparse
from receptor_utils import simple_bio_seq as simple
import urllib.request
import io


def get_parser():
    parser = argparse.ArgumentParser(description='Extract reference files for nominated species from a gapped IMGT reference file')
    parser.add_argument('species_name', help='species name for IMGT file in quotes, e.g. "Homo sapiens", "Macaca mulatta"')
    parser.add_argument('-r', '--ref', help='gapped imgt reference file. If not specified, the current file will be downloaded from http://www.imgt.org/download/GENE-DB/IMGTGENEDB-ReferenceSequences.fasta-nt-WithGaps-F+ORF+inframeP')
    parser.add_argument('-L', '--locus', help='locus identifier, e.g. "IGH" (default), "IGK", "TRB", "TRG"', default='IGH', required=False)
    parser.add_argument('-F', '--functional_only', action='store_true')
    return parser

def main():
    args = get_parser().parse_args()

    imgt_url = "https://www.imgt.org/download/GENE-DB/IMGTGENEDB-ReferenceSequences.fasta-nt-WithGaps-F+ORF+inframeP"

    segs = {
        'IGH': ['IGHV', 'IGHD', 'IGHJ', 'CH'],
        'IGK': ['IGKV', 'IGKJ', 'IGKC'],
        'IGL': ['IGLV', 'IGLJ', 'IGLC'],
        'TRA': ['TRAV', 'TRAJ', 'TRAC'],
        'TRB': ['TRBV', 'TRBD', 'TRBJ', 'TRBC'],
        'TRG': ['TRGV', 'TRGJ', 'TRGC'],
        'TRD': ['TRDV', 'TRDD', 'TRDJ', 'TRDC']
    }

    if args.ref:
        refs = simple.read_imgt_fasta(args.ref, [args.species_name], segs[args.locus], functional_only=args.functional_only)
    else:
        with urllib.request.urlopen(imgt_url) as fi:
            refs = simple.read_imgt_fasta(io.StringIO(fi.read().decode('utf-8')), [args.species_name], segs[args.locus], functional_only=args.functional_only)

    simple.write_fasta('%s_IGHV_gapped.fasta'.replace('IGHV', args.locus + 'V') %args.species_name.replace(' ', '_'), refs[args.species_name][args.locus + 'V'])

    ungapped = {}

    for seg in segs[args.locus]:
        ungapped[seg] = {}
        for id, seq in refs[args.species_name][seg].items():
            ungapped[seg][id] = seq.replace('.', '')

        simple.write_fasta('%s_%s.fasta' % (args.species_name.replace(' ', '_'), seg), ungapped[seg])


if __name__ == "__main__":
    main()



#!python
# Make igblast ndm file from a gapped reference set

# Copyright (c) 2021 William Lees

# This source code, and any executable file compiled or derived from it, is governed by the European Union Public License v. 1.2,
# the English version of which is available here: https://perma.cc/DK5U-NDVE


import csv
from receptor_utils import simple_bio_seq as simple
import argparse
import json

def get_parser():
    parser = argparse.ArgumentParser(description='Make IMGT-delineated igblast ndm file from an IMGT-gapped variable-region reference set or AIRR-C JSON format file.')
    parser.add_argument('ref_file', help='gapped reference set (.fasta, .json)')
    parser.add_argument('chain', help='chain as required by igblast (eg VH)')
    parser.add_argument('ndm_file', help='output file (ndm)')
    parser.add_argument('--cdr_coords', '-c', help='comma-separated list of 1-based co-ordinates for CDR1, CDR2 and CDR3-start (default 79,114,166,195,313)')
    return parser


def main():
    args = get_parser().parse_args()

    allowed_chains = ['VH', 'VK', 'VL', 'VA', 'VB', 'VD', 'VG']

    if args.chain not in allowed_chains:
        print(f"Error: chain must be one of {', '.join(allowed_chains)}")
        exit(1)

    cdr_coords = [79, 114, 166, 195, 313]
    if args.cdr_coords is not None:
        if not args.ref_file.lower().endswith('.fasta'):
            print("Error: CDR coordinates are only used with fasta files")
            exit(1)
        cdr_coords = args.cdr_coords.split(',')
        if len(cdr_coords) != 5:
            print("Error: CDR coordinates must be a comma-separated list of 5 integers")
            exit(1)
        try:
            cdr_coords = [int(x) for x in cdr_coords]
        except:
            print("Error: CDR coordinates must be a comma-separated list of 5 integers")
            exit(1)
        
        last_coord = 0
        for coord in cdr_coords:
            if coord < 0 or coord <= last_coord:
                print("Error: CDR coordinates must be positive integers in ascending order")
                exit(1)
            last_coord = coord
    
    if args.ref_file.lower().endswith('.json'):
        from_json(args)
    elif args.ref_file.lower().endswith('.fasta'):
        from_fasta(args, cdr_coords)
    else:
        print("Error: reference file must be a .json or .fasta file")
        exit(1)


def from_fasta(args, cdr_coords):
    seqs = simple.read_fasta(args.ref_file)
    
    
    # desired headers:
    # gene/allele name, FWR1 start, FWR1 stop, CDR1 start, CDR1 stop, FWR2 start, FWR2 stop, CDR2 start, CDR2 stop, FWR3 start, FWR3 stop, chain type, coding frame start.
    # FWR/CDR positions are 1-based while the coding frame start positions are 0-based
    
    omissions = False

    with open(args.ndm_file, 'w') as fo:
        for label, seq in seqs.items():

            # If the sequence is truncated at the 5' end, don't include the sequence

            if seq[0] == '.':
                print(f"Omitting allele {label} as it is truncated at the 5' end")
                omissions = True
                continue

            fo.write(record_from_gapped_seq(label, seq, args.chain, cdr_coords))

    if omissions:
        print('Omissions will not affect results provided at least one non-truncated allele of the gene is present.')


def record_from_gapped_seq(label, seq, chain, cdr_coords):
    # Python-based ranges of IMGT elements
    # default: 79,114,166,195,313
    # <---------------------------------- FR1-IMGT -------------------------------->______________ CDR1-IMGT ___________<-------------------- FR2-IMGT ------------------->___________ CDR2-IMGT ________<----------------------------------------------------- FR3-IMGT ----------------------------------------------------> CDR3-IMGT
    imgt_fr1 = (0, cdr_coords[0] - 1)
    imgt_cdr1 = (cdr_coords[0] - 1, cdr_coords[1])
    imgt_fr2 = (cdr_coords[1], cdr_coords[2] - 1)
    imgt_cdr2 = (cdr_coords[2] - 1, cdr_coords[3])
    imgt_fr3 = (cdr_coords[3], cdr_coords[4] - 1)

    rec = []
    rec.append(label)
    pos = 1
    rec.append(str(pos))     # FWR start
    pos += len(seq[slice(*imgt_fr1)].replace('.', '')) - 1
    rec.append(str(pos))    # FWR1 stop
    pos += 1
    rec.append(str(pos))     # CDR1 start
    pos += len(seq[slice(*imgt_cdr1)].replace('.', '')) - 1
    rec.append(str(pos))     # CDR1 end
    pos += 1
    rec.append(str(pos))     # FWR2 start
    pos += len(seq[slice(*imgt_fr2)].replace('.', '')) - 1
    rec.append(str(pos))     # FWR2 end
    pos += 1
    rec.append(str(pos))     # CDR2 start
    pos += len(seq[slice(*imgt_cdr2)].replace('.', '')) - 1
    rec.append(str(pos))     # CDR2 end
    pos += 1
    rec.append(str(pos))     # FWR3 start
    pos += len(seq[slice(*imgt_fr3)].replace('.', '')) - 1
    rec.append(str(pos))     # FWR3 end
    rec.append(chain)    # chain type
    rec.append('0')       # coding frame start
    return '\t'.join(rec) + '\n'
    


def from_json(args):
    with open(args.ref_file, 'r') as fo:
        ref = json.load(fo)

    if isinstance(ref["GermlineSet"], list):
        germline_set = ref["GermlineSet"][0]
    else:
        germline_set = ref["GermlineSet"]

    with open(args.ndm_file, 'w') as fo:
        for allele in germline_set['allele_descriptions']:
            if allele['sequence_type'] == 'V':
                if 'v_gene_delineations' not in allele:
                    print(f"Omitting allele {allele['label']} as no delineations are specified")
                    continue

                for delineation in allele['v_gene_delineations']:
                    if delineation['delineation_scheme'] == 'IMGT':
                        for coord in ['fwr1_start', 'fwr1_end', 'cdr1_start', 'fwr2_end', 'cdr2_start', 'fwr3_end']:
                            if coord not in delineation or delineation[coord] is None:
                                print(f"Omitting incomplete allele {allele['label']}")
                                continue

                        rec = record_from_json(allele['label'], delineation, args.chain)
                        
                        # compare with gapped sequence derivation and warn if there is a mismatch
                        # can't make this comparison any longer because we don't know the coordinates used in the gapped alignment
                        # rec_from_gapped = record_from_gapped_seq(allele['label'], delineation['aligned_sequence'], args.chain)

                        # if rec != rec_from_gapped:
                        #     print(f"Warning: mismatch between IMGT gapped sequence and CDR delineation for allele {allele['label']}")

                        fo.write(rec)

    


def record_from_json(label, d, chain):
    return(f"{label}\t{d['fwr1_start']}\t{d['fwr1_end']}\t{d['cdr1_start']}\t{d['cdr1_end']}\t{d['fwr2_start']}\t{d['fwr2_end']}\t{d['cdr2_start']}\t{d['cdr2_end']}\t{d['fwr3_start']}\t{d['fwr3_end']}\t{chain}\t0\n")


if __name__ == '__main__':
    main()  



# Merge two fasta files, keeping records in file 2 that do not match sequences in file 1

import argparse
from receptor_utils import simple_bio_seq as simple


def get_parser():
    parser = argparse.ArgumentParser(description='Merge two fasta files, keeping records in file 2 that do not match sequences in file 1')
    parser.add_argument('file_1', help='first file to merge')
    parser.add_argument('file_2', help='second file to merge')
    parser.add_argument('outfile', help='output file')
    return parser


def main():
    args = get_parser().parse_args()

    seqs = simple.read_fasta(args.file_1)
    seqs_map = {v: k for k, v in seqs.items()}

    merge_seqs = simple.read_fasta(args.file_2)

    added = 0

    for m_name, m_seq in merge_seqs.items():
        if m_seq not in seqs_map:
            seqs[m_name] = m_seq
            added += 1

    simple.write_fasta(args.outfile, seqs)
    print('%d items merged from %s.' % (added, args.file_2))


if __name__ == "__main__":
    main()



# Copyright (c) 2022 William Lees

# This source code, and any executable file compiled or derived from it, is governed by the European Union Public License v. 1.2,
# the English version of which is available here: https://perma.cc/DK5U-NDVE


from receptor_utils import simple_bio_seq as simple
import argparse


def get_parser():
    parser = argparse.ArgumentParser(description='Return the reverse complement of a nucleotide sequence')
    parser.add_argument('sequence', help='input sequence')
    return parser


def main():
    args = get_parser().parse_args()

    rc = simple.reverse_complement(args.sequence)
    print(rc)



from receptor_utils import simple_bio_seq as simple

seqs = [
    'abcdef',
    'abcded',
    'aacdef',
]

seq_dict = {}
for i in range(len(seqs)):
    seq_dict[str(i)] = seqs[i]

c, s = simple.scored_consensus(seq_dict)

breakpoint()


from warnings import warn
from . import number_v

def __getattr__(name):
    if name == 'number_ighv':
        warn("number_ighv is deprecated. Please use number_v instead.", DeprecationWarning, stacklevel=2)
        return number_v
   
    raise AttributeError(f"No module named {name}")




# Provide a name for a novel allele sequence, using the VDJbase approach as outlined
# at https://wordpress.vdjbase.org/index.php/vdjbase_help/airr-seq-data-allele-names/
# For V sequences, the sequence must be full-length at the 5' end, or IMGT-gapped.

# Copyright (c) 2021 William Lees

# This source code, and any executable file compiled or derived from it, is governed by the European Union Public License v. 1.2,
# the English version of which is available here: https://perma.cc/DK5U-NDVE


from receptor_utils import simple_bio_seq as simple
from receptor_utils import number_v
from collections import namedtuple
from Bio import Align


# Make a name for the novel allele, given its gapped or ungapped sequence
# returns a tuple (name, gapped_sequence)
#
# novel_seq: the full-length sequence to name, which may be either gapped or ungapped in the case of a V-gene
# ref_set: dict of reference genes (gapped in the case of v-genes), in the format returned by read_fasta
# v_gene: True if we are naming a v_gene
# gaps in V-sequences are '.', as in the IMGT reference set


def name_novel(novel_seq: str, ref_set: dict, v_gene: bool = True):

    """Make a name for the novel allele, given its gapped or ungapped sequence.
    The name conforms to the description `here <https://wordpress.vdjbase.org/index.php/vdjbase_help/airr-seq-data-allele-names/>`_ .
    The sequence must be full-length at the 5 prime end, or gapped

    :param novel_seq: the full-length sequence to name, which may be either gapped or ungapped in the case of a V-gene
    :type novel_seq: str
    :param ref_set: dict of reference genes (gapped in the case of v-genes), in the format returned by read_fasta
    :type ref_set: dict
    :param v_gene: True only if we are naming a v_gene
    :type v_gene: bool
    :return: tuple consisting of three strings: novel_name, novel_seq, notes. novel_seq is gapped in the case of a V-gene
    :rtype: tuple
    """

    novel_seq = novel_seq.upper()
    notes = ''

    # Check for truncation at 5' end and construct ambiguous reference set if necessary

    if novel_seq[0] == '.':
        for i in range(0, len(novel_seq)):
            if novel_seq[i] != '.':
                break
        ref_set = build_ambiguous_ref(ref_set, i)

        # TODO - handle ungapped sets that are truncated at the 5' end
    else:
        # gap the sequence (even if already gapped) so that we get the notes
    
        if v_gene:
            if '.' in novel_seq:
                novel_seq = novel_seq.replace('.', '')
    
            ungapped_ref_set = {}
            for k, v in ref_set.items():
                ungapped_ref_set[k] = v.replace('.', '')
            novel_seq, aa, notes = number_v.gap_sequence(novel_seq, ref_set, ungapped_ref_set)


    closest_ref_name = closest_aligned_ref(novel_seq, ref_set)[0]
    closest_ref_seq = ref_set[closest_ref_name]

    diffs = {}
    insertions = []
    class Diff:
        pass

    # check for indels

    n_diffs = simple.nt_diff(closest_ref_seq, novel_seq)
    if n_diffs > 0:
        ungapped_closest_ref = closest_ref_seq.replace('.', '')
        fixed_seq, insertions = find_indels(ungapped_closest_ref, novel_seq.replace('.', ''))
        fixed_diffs = simple.nt_diff(ungapped_closest_ref, fixed_seq)
        if fixed_diffs < n_diffs:
            # thread the gaps through the fixed sequence and adjust insertion positions
            novel_seq = ''
            fixed_ind = 0
            for i in range(len(closest_ref_seq)):
                if closest_ref_seq[i] != '.':
                    novel_seq += fixed_seq[fixed_ind]
                    fixed_ind += 1
                else:
                    novel_seq += '.'
                    for insertion in insertions:
                        if insertion[0] > i:
                            insertion[0] += 1
                            
            if fixed_ind < len(fixed_seq):
                for i in range(fixed_ind, len(fixed_seq)):
                    novel_seq += fixed_seq[i]
                            
    if len(closest_ref_seq) < len(novel_seq):
        closest_ref_seq += '.' * (len(novel_seq) - len(closest_ref_seq))

    for i in range(0, len(novel_seq)):
        if closest_ref_seq[i] != novel_seq[i]:
            d = Diff()
            d.pos = i
            d.ref = closest_ref_seq[i]
            d.novel = novel_seq[i]
            diffs[i] = d

    # trim any diffs that are past the end of the novel sequence

    for i in range(len(novel_seq) - 1, -1, -1):
        if novel_seq[i] != '.':
            break

    for i in list(diffs.keys()):
        if i >= len(novel_seq):
            del diffs[i]

    # add diffs if the closest ref is longer than the novel seq

    if len(closest_ref_seq) > len(novel_seq):
        for j in range(len(novel_seq), len(closest_ref_seq)):
            d = Diff()
            d.pos = j
            d.ref = closest_ref_seq[j]
            d.novel = '-'
            diffs[j] = d


    # consolidate any adjacent diffs

    prev_i = None
    for i, d in list(diffs.items()):
        if prev_i and diffs[prev_i].pos + len(diffs[prev_i].novel) == i:
            diffs[prev_i].ref += closest_ref_seq[i]
            diffs[prev_i].novel += novel_seq[i] if i < len(novel_seq) else '-'
            del diffs[i]
        else:
            prev_i = i

    # split diffs that are too short to consolidate

    for i, d in list(diffs.items()):
        if 1 < len(d.novel) < 4 and '.' not in d.novel and '.' not in d.ref:
            for j in range(1, len(d.novel)):
                diffs[i + j] = Diff()
                diffs[i + j].pos = d.pos + j
                diffs[i + j].ref = d.ref[j]
                diffs[i + j].novel = d.novel[j]
            d.ref = d.ref[0]
            d.novel = d.novel[0]

    # form the suffix

    all_positions = list(diffs.keys())
    all_positions.extend([i[0] for i in insertions])
    all_positions = list(set(all_positions))
    all_positions.sort()

    suffs = []

    # create the name suffix parts
    
    for i in all_positions:
        if i in diffs:
            d = diffs[i]
            if len(d.novel) == 1 and d.ref != '.' and d.novel != '.':
                suffs.append('%s%d%s' % (d.ref.lower(), d.pos + 1, d.novel.lower()))
            else:
                suffs.append('%d%s%d' % (d.pos + 1, d.novel.lower(), d.pos + len(d.novel)))

        insertions_at_i = [insertion for insertion in insertions if insertion[0] == i]
        if len(insertions_at_i) > 0:
            suffs.append('i%d%s' % (i + 1, ''.join([insertion[1].lower() for insertion in insertions_at_i])))

    novel_name = closest_ref_name + '_' + '_'.join(suffs) if len(suffs) else closest_ref_name

    # put the insertions into the sequence. 

    for i in range(len(insertions) - 1, -1, -1):
        novel_seq = novel_seq[:insertions[i][0] + 1] + insertions[i][1] + novel_seq[insertions[i][0] + 1:]

    return novel_name, novel_seq, notes


def aligned_diff(novel_seq: str, ref_seq: str):
    aligner = Align.PairwiseAligner()
    aligner.mode = 'global'
    aligner.open_gap_score = -5
    aligner.extend_gap_score = -1
    aligner.match_score = 2
    aligner.mismatch_score = 0
    aligner.target_end_gap_score = 1

    best_alignment = aligner.align(novel_seq, ref_seq)[0]

    # return 0-based start, end and score
    return best_alignment[0], best_alignment[1], best_alignment.score


# Find the closest reference sequence to a query sequence using a global alignment
# Returns a list of equally close reference names
def closest_aligned_ref(seq: str, ref: dict):
    """Given an input sequence, find the closest entry or entries in an IMGT or other reference set as determined by a global alignment.

    :param seq: the input sequence
    :type seq: str
    :param ref: the reference set
    :type ref: dict
    :return: the name of the closest entry, entries, if >1 sequences in the reference set were equally close
    :rtype: list
    """

    scores = [(seq_name, aligned_diff(seq, ref_seq)[2]) for seq_name, ref_seq in ref.items()]
    max_score = max(scores, key=lambda x: x[1])[1]
    closest_names = [seq_name for seq_name, score in scores if score == max_score]

    return closest_names


# Align novel sequence against the reference sequence and check for indels
def find_indels(closest_ref_seq, novel_seq):
    insertions = []
    fixed_novel = ''        # this will be the novel without any insertions, and now aligned so that there is a '-' for any deletion

    seqA, seqB, _ = aligned_diff(closest_ref_seq, novel_seq)
    seqA = list(seqA)

    for i in range(len(seqA)-1, -1, -1):
        if seqA[i] == '-':
            del seqA[i]
        else:
            break
    seqA = ''.join(seqA)

    for c in zip(list(seqA), list(seqB)):
        if c[0] != '-':
            fixed_novel += c[1]
        else:
            insertions.append([len(fixed_novel) - 1, c[1]])

    if len(seqB) > len(seqA):
        for i in range(len(seqA), len(seqB)):
            fixed_novel += seqB[i]

    return fixed_novel, insertions


# Build (and name) a reference set for sequences truncated at the 5' end
# Start is 0-based index of first available bp


def build_ambiguous_ref(full_ref, start):
    rev_ref = {}
    for name, seq in full_ref.items():
        seq = '.' * start + seq[start:]
        if seq not in rev_ref:
            rev_ref[seq] = []
        rev_ref[seq].append(name)

    ref = {}
    for seq, names in rev_ref.items():
        if len(names) == 1:
            ref[names[0]] = seq
        else:
            genes = {}
            for name in names:
                gene, allele = name.split('*')
                if gene not in genes:
                    genes[gene] = []
                genes[gene].append(allele)
            if len(genes) == 1:
                name = gene + '*' + '_'.join(sorted(genes[gene]))
                ref[name] = seq
            else:
                gene_list = sorted(genes.keys())
                name = gene_list[0] + '*' + '_'.join(sorted(genes[gene_list[0]]))
                fam_0 = gene_list[0].split('-')[0]
                for i in range(1, len(gene_list)):
                    name += '_'
                    fam = gene_list[i].split('-')[0]
                    num = gene_list[i].replace(fam + '-', '')
                    if fam == fam_0:
                        alleles = [num + '.' + a for a in genes[gene_list[i]]]
                        name += '_'.join(alleles)
                    else:
                        fam = gene_list[i][4:]
                        alleles = [fam + '.' + a for a in genes[gene_list[i]]]
                        name += '_'.join(alleles)
                ref[name] = seq

    return ref


def run_tests():
    ref_set = simple.read_fasta('Homo_sapiens_IGHV_gapped.fasta')
    test_novels = simple.read_fasta('IGHV3-20.04.fasta')

    for name, seq in test_novels.items():
        novel_name = name_novel(seq, ref_set)[0]
        if novel_name != name:
            print('Error: %s != %s' % (novel_name, name))


if __name__ == "__main__":
    run_tests()




#!python
# Annotate a set of J sequences, finding the frame alignment and index of the first nucleotide of the conserved PHE or TRP

# Copyright (c) 2021 William Lees

# This source code, and any executable file compiled or derived from it, is governed by the European Union Public License v. 1.2,
# the English version of which is available here: https://perma.cc/DK5U-NDVE


import csv
from receptor_utils import simple_bio_seq as simple
import argparse
import re
from collections import namedtuple
import json


def main():
    args = get_parser().parse_args()

    if args.seq_file.lower().endswith('.json'):
        seqs = seqs_from_json(args)
    elif args.seq_file.lower().endswith('.fasta'):
        seqs = simple.read_fasta(args.seq_file)
    else:
        print("Error: reference file must be a .json or .fasta file")
        exit(1)

    
    annotations = []
    
    for seq_name, seq in seqs.items():
        solutions = []
        Result = namedtuple('Result', ['nt_seq', 'frame', 'cdr3_pos', 'trans', 'extra_bps'])
        for frame in range(3):
            trans = simple.translate(seq[frame:])

            if '*' in trans:
                continue

            for m in re.finditer('[WF]G.G', trans):
                solutions.append(Result._make([
                    seq,
                    frame,
                    frame + 3*m.start(),
                    trans,
                    (len(seq) - frame) % 3
                ]))

        if solutions:
            if len(solutions) > 1:
                print('WARNING: multiple solutions for {seq_name}')
                
            for solution in solutions:
                print(f'{seq_name}: j_codon_frame {solution.frame} j_cdr3_end {solution.cdr3_pos-1} exta bps {solution.extra_bps}')

                loc_count = '      '
                for num in range(len(solution.trans)):
                    loc_count += f' {num+1:<2}'
                print(loc_count)

                aa = '      '
                for num in range(len(solution.trans)):
                    aa += f' {solution.trans[num]} '
                print(aa)

                nt = f'  {solution.nt_seq[:solution.frame]:>3} {solution.nt_seq[solution.frame:]}'
                print(nt + '\n\n')

                annotations.append({
                    '#name': seq_name,
                    'j_codon_frame': solution.frame,
                    'chain_type': f'{seq_name[3]}{seq_name[2]}',
                    'j_cdr3_end': solution.cdr3_pos-1,
                    'extra_bps': solution.extra_bps,
                })
        else:
            print(f'{seq_name}: no solutions')

    simple.write_csv(args.out_file, annotations, delimiter='\t')


def get_parser():
    parser = argparse.ArgumentParser(
        description='Annotate a set of J sequences, finding the frame alignment and 1-based index of the first nucleotide of the conserved PHE or TRP')
    parser.add_argument('seq_file', help='J sequences to annotate (fasta or AIRR-C JSON format)')
    parser.add_argument('out_file', help='Annotation output in aux format for use with IgBlast')
    return parser


def seqs_from_json(args):
    with open(args.seq_file, 'r') as fo:
        ref = json.load(fo)

    if isinstance(ref["GermlineSet"], list):
        germline_set = ref["GermlineSet"][0]
    else:
        germline_set = ref["GermlineSet"]

    seqs = {}

    for allele in germline_set['allele_descriptions']:
        if allele['sequence_type'] == 'J':
            seqs[allele['label']] = allele['coding_sequence']

    return seqs


if __name__ == "__main__":
    main()




# Copyright (c) 2021 William Lees

# This source code, and any executable file compiled or derived from it, is governed by the European Union Public License v. 1.2,
# the English version of which is available here: https://perma.cc/DK5U-NDVE

import re
from receptor_utils import simple_bio_seq as simple

def nt_diff(s1, s2):
    diffs = 0
    row = list(zip(s1.upper(), s2.upper()))

    for p in range(len(row)):
        if row[p][0] != row[p][1] and (row[p][0] not in ['X', 'N'] and row[p][1] not in ['X', 'N']):
            diffs += 1

    diffs += abs(len(s1) - len(s2))
    return diffs


# De novo IMGT-gapping of IGHV germline genes, using guidance from http://www.bioinf.org.uk/abs/info.html#cdrid to confirm CDR1 and 2
# This is experimental. Not used in the examples and specific to IGHV. Maybe a recourse if there is no suitable gapped reference template
# to rely on
def number_ighv(seq):
    gapped = ''
    # find the conserved Cys at 23
    if 'C' not in seq[21:23]:
        return None, 'First conserved cysteine not found'

    if seq[21] == 'C':
        gapped += seq[:9]
        gapped += '.'
        gapped += seq[9:25]
        next_p = 25
    else:
        gapped += seq[:26]
        next_p = 26

    # find the conserved Trp

    if 'W' not in seq[33:42]:
        return None, 'Conserved Trp not found'

    trp_pos = [pos.start()+33 for pos in re.finditer('W', seq[33:42])]

    max_conf = -1
    best_result = ''
    for pos in trp_pos:
        (confidence, ret) = number_from_trp(seq, next_p, gapped, pos)
        if confidence is not None and confidence > max_conf:
            best_result = ret
            max_conf = confidence

    if max_conf >= 0:
        return best_result, ''
    else:
        return None, 'CDR2 or 2nd Cys not found'


def number_from_trp(seq, next_p, gapped, trp_pos):
    # add the CDR1 residues
    gapped += distribute_cdr(seq[next_p:trp_pos-2], 12)
    gapped += seq[trp_pos-2:trp_pos+1]
    next_p = trp_pos+1

    # allocate FR2 residues
    gapped += seq[next_p:next_p + 14]
    next_p = next_p + 14        # seq[next_p] is now the first aa of cdr2

    # look for motif trailing cdr2 (should start at position 75)

    found_motif = False
    motif_range = seq[next_p+2+8:next_p+10+8+3]    # allow for cdr2 min length 2, max length 10 and motif length 3
    for match in match_score(motif_range, [['K', 'R'], ['L', 'I', 'V', 'F', 'T', 'A'], ['T', 'S', 'I', 'A']], 2):
        # check for conserved cysteine at position 104
        cys_range = seq[next_p+2+10+match[0]+26:next_p+2+10+match[0]+29]
        if 'C' in cys_range:
            c_index = cys_range.index('C')
            match = next_p + 2 + 10 + match[0]     # now an absolute index of the motif in seq
            found_motif = True
            break

    if not found_motif:
        return (None, None)

    # Build a confidence score based on leading residue sequence
    cdr2_leader = gapped[49:54]
    confidence = match_score(cdr2_leader, [['L'], ['E'], ['W'], ['I'], ['G']], 0)
    if len(confidence) > 0:
        confidence = confidence[0][1]  # only one result, because string lengths are the same

    # add the CDR2 residues
    cdr2 = seq[next_p:match - 10]
    gapped += distribute_cdr(cdr2, 10)

    # add the remaining v-region residues
    gapped += seq[match - 10:]

    # insert gaps in FR3
    if c_index < 2:
        gaps = '.' * (2 - c_index)
        gapped = gapped[:72] + gaps + gapped[72:]

    return (confidence, gapped)


def match_score(target, match_list, thresh):
    res = []
    for i in range(len(target)-len(match_list) + 1):
        score = 0
        for j in range(len(match_list)):
            if target[i+j] in match_list[j]:
                score += 1

        if score >= thresh:
            res.append((i, score))

    return res

# distribute aas in the cdr according to the numbering scheme
def distribute_cdr(cdr, length):
    cdr_div = int(len(cdr)/2)

    if 2 * cdr_div == len(cdr):
        return cdr[:cdr_div] + '.' * (length - len(cdr)) + cdr[cdr_div:]
    else:
        return cdr[:cdr_div+1] + '.' * (length - len(cdr)) + cdr[cdr_div+1:]

pretty_header = """
           FR1-IMGT          CDR1-IMGT       FR2-IMGT     CDR2-IMGT                 FR3-IMGT
            (1-26)            (27-38)        (39-55)       (56-65)                  (66-104)
1       10        20         30         40        50         60         70        80        90        100     
.........|.........|...... ...|........ .|.........|..... ....|..... ....|.........|.........|.........|...."""


# insert space before the position pos (zero numbering)
def insert_space(seq, pos):
    return seq[:pos] + ' ' + seq[pos:]

# pretty print of gapped sequence
def pretty_gapped(seq):
    seq = insert_space(seq, 65)
    seq = insert_space(seq, 55)
    seq = insert_space(seq, 38)
    seq = insert_space(seq, 26)

    print(pretty_header)
    print(seq)


# Given a gapped AA sequence and an ungapped nt sequence, produce a gapped nt sequence
def gap_nt_from_aa(nucleotide_seq, peptide_seq):
    codons = [codon for codon in simple.chunks(nucleotide_seq, 3)]  #splits nucleotides into codons (triplets)
    remains = nucleotide_seq[3*len(codons):] if 3*len(codons) != len(nucleotide_seq) else ''
    gapped_codons = []
    codon_count = 0

    for aa in peptide_seq:  #adds '---' gaps to nucleotide seq corresponding to peptide
        if aa not in ('-', '.'):
            gapped_codons.append(codons[codon_count])
            codon_count += 1
        else:
            gapped_codons.append('...')

    while codon_count < len(codons):                # account for trailing nt, eg fragment of a codon
        gapped_codons.append(codons[codon_count])
        codon_count += 1

    return ''.join(gapped_codons) + remains


def gap_align(seq, ref):
    seq_i = iter(seq)

    res = ''
    started = False

    for r in ref:
        if r != '.':
            started = True
            try:
                res += next(seq_i)
            except StopIteration:
                res += '.'
        else:
            if started:
                res += '.'
            else:
                try:
                    res += next(seq_i)
                except StopIteration:
                    res += '.'

    while True:
        try:
            res += next(seq_i)
        except StopIteration:
            break

    # remove trailing gaps

    i = len(res) - 1

    while i and res[i] == '.':
        i -= 1

    res = res[:i+1]


    return res


def gap_align_aa(seq, ref):
    seq_i = iter(seq)

    res = ''
    started = False

    for r in ref:
        if r != '.':
            started = True
            try:
                res += next(seq_i)
            except StopIteration:
                break
        else:
            if started:
                res += '.'
            else:
                try:
                    res += next(seq_i)
                except StopIteration:
                    break

    while True:
        try:
            res += next(seq_i)
        except StopIteration:
            break

    return res


def gap_align_aa_from_nt(aa_seq, nt_gapped):
    aa_i = iter(aa_seq)
    gapped_aa = ''

    started = False

    for codon in simple.chunks(nt_gapped, 3):
        if '.' in codon:
            if started:
                gapped_aa += '.'
            else:
                try:
                    gapped_aa += next(aa_i)
                except StopIteration:
                    break

        else:
            started = True

            try:
                gapped_aa += next(aa_i)
            except StopIteration:
                break

    return gapped_aa


# gap using the closest reference
def gap_sequence(seq, gapped_ref, ungapped_ref):
    differences = [(k, nt_diff(seq, v)) for k, v in ungapped_ref.items()]
    closest = min(differences, key=lambda x: x[1])[0]

    # print(f'gapping with {closest}')
    res = gap_align(seq, gapped_ref[closest])
    aa = simple.translate(seq)
    aa = gap_align_aa_from_nt(aa, res)

    notes = ''
    # checks
    if 'V' in closest:
        notes = check_conserved_residues(aa)

    return (res, aa, notes)


def check_conserved_residues(aa):
    notes = []

    if len(aa) < 104:
        return 'Sequence truncated before second cysteine'

    # allow stop codon in final nucleotides without warning
    if '*' in aa:
        pos = aa.index('*')
        if pos <= 103:
            notes.append('Stop codon in V-REGION before 2nd cysteine')

    if aa[22] != 'C':
        notes.append('First cysteine not found')

    if aa[40] != 'W':
        notes.append('Conserved Trp not found')

    if aa[103] != 'C':
        notes.append('Second cysteine not found')

    return ', '.join(notes)


def run_tests():
    failed = 0

    print('1-64')
    query = 'EMQLVQSEAEVKKPGASVKISCKASGYTFTYRYLHWLRQTPGQGLEWMGWITPYNGNTNYAQKFQDRATITRDRSMSTAYMELSSLRSEDTAVYYCAR'

    (seq, status) = number_ighv(query)

    if seq is not None:
        pretty_gapped(seq)
        if seq == 'EMQLVQSEA.EVKKPGASVKISCKASGYTF....TYRYLHWLRQTPGQGLEWMGWITPY..NGNTNYAQKFQ.DRATITRDRSMSTAYMELSSLRSEDTAVYYCAR':
            print('passed')
        else:
            print('FAILED')
            failed += 1
    else:
        print(query + '->')
        print(status)
        failed += 1


    print('2 -143')
    query = 'QVTLKESGPALVKPTQTLTLTCTFSGFSLTTSGMGVGWIRQPPGKALEWLALIYWDDDKRYSTSLKSRLTISKDTSKNQVVLTMTNMDPMDTATYYCARG'

    (seq, status) = number_ighv(query)

    if seq is not None:
        pretty_gapped(seq)
        if seq == 'QVTLKESGP.ALVKPTQTLTLTCTFSGFSLT..TSGMGVGWIRQPPGKALEWLALIYWD...DDKRYSTSLK.SRLTISKDTSKNQVVLTMTNMDPMDTATYYCARG':
            print('passed')
        else:
            print('FAILED')
            failed += 1
    else:
        print(query + '->')
        print(status)
        failed += 1

    print('3-8')
    query = 'EVQLVESGGGLVQPGRSLRPSCAASGFTFSSYGMHWVRQAPEEGLVWVSYIGSSTMYYADSVKGRFTISRDNAKNSLYLQMNSLRAEDTAVYYCVR'

    (seq, status) = number_ighv(query)

    if seq is not None:
        pretty_gapped(seq)
        if seq == 'EVQLVESGG.GLVQPGRSLRPSCAASGFTF....SSYGMHWVRQAPEEGLVWVSYIGS....STMYYADSVK.GRFTISRDNAKNSLYLQMNSLRAEDTAVYYCVR':
            print('passed')
        else:
            print('FAILED')
            failed += 1
    else:
        print(query + '->')
        print(status)
        failed += 1

    print('4-86')
    query = 'QVQLQESGPGLVKPSETLSLTCAVSGGSISSSNWWSWIRQPPGKGLEWIGRISGSGGSTSDNPSLKSRVTISKDTSKNQFSLKLSSVTAADTAVYYCAR'

    (seq, status) = number_ighv(query)

    if seq is not None:
        pretty_gapped(seq)
        if seq == 'QVQLQESGP.GLVKPSETLSLTCAVSGGSIS...SSNWWSWIRQPPGKGLEWIGRISGS..GGSTSDNPSLK.SRVTISKDTSKNQFSLKLSSVTAADTAVYYCAR':
            print('passed')
        else:
            print('FAILED')
            failed += 1
    else:
        print(query + '->')
        print(status)
        failed += 1

    print('5-16')
    query = 'EVQLVQSGAEVKRPGESLKISCKTSGYSFTSYWISWVRQMPGKGLEWMGAIDPSDSDTRYNPSFQGQVTISADKSISTAYLQWSRLKASDTATYYCAK'

    (seq, status) = number_ighv(query)

    if seq is not None:
        pretty_gapped(seq)
        if seq == 'EVQLVQSGA.EVKRPGESLKISCKTSGYSF....TSYWISWVRQMPGKGLEWMGAIDPS..DSDTRYNPSFQ.GQVTISADKSISTAYLQWSRLKASDTATYYCAK':
            print('passed')
        else:
            print('FAILED')
            failed += 1
    else:
        print(query + '->')
        print(status)
        failed += 1

    print('6-1')
    query = 'QVQLQESGPGLVKPSQTLSLTCAISGDSVSSNSATWNWIRQSPSRGLEWLGRTYYRSKWYNDYAQSVQNRISINPDTSKNQFSLQLNSVTPEDMAVYYCAR'

    (seq, status) = number_ighv(query)

    if seq is not None:
        pretty_gapped(seq)
        if seq == 'QVQLQESGP.GLVKPSQTLSLTCAISGDSVS..SNSATWNWIRQSPSRGLEWLGRTYYRS.KWYNDYAQSVQ.NRISINPDTSKNQFSLQLNSVTPEDMAVYYCAR':
            print('passed')
        else:
            print('FAILED')
            failed += 1
    else:
        print(query + '->')
        print(status)
        failed += 1

    print('7-108')
    query = 'QVQLVQSGAEVKQPGASVKVSCKASGYTFTSYGMNWVRQAHGQRLEWMGWINTDTGNPTYAQGFKERFTFSMDTSISTAYLQISSLKAEDTAVYYCAR'

    (seq, status) = number_ighv(query)

    if seq is not None:
        pretty_gapped(seq)
        if seq == 'QVQLVQSGA.EVKQPGASVKVSCKASGYTF....TSYGMNWVRQAHGQRLEWMGWINTD..TGNPTYAQGFK.ERFTFSMDTSISTAYLQISSLKAEDTAVYYCAR':
            print('passed')
        else:
            print('FAILED')
            failed += 1
    else:
        print(query + '->')
        print(status)
        failed += 1


    if failed == 0:
        print('All tests passed.')
    else:
        print('%d tests failed.' % failed)


if __name__ == "__main__":
    run_tests()



# Copyright (c) 2022 William Lees

# This source code, and any executable file compiled or derived from it, is governed by the European Union Public License v. 1.2,
# the English version of which is available here: https://perma.cc/DK5U-NDVE

# Name an allele by comparing to a reference set. Construct a novel name if necessary

from receptor_utils import simple_bio_seq as simple
from receptor_utils import novel_allele_name
import argparse


def find_gene(name):
    gene = name
    if '*' in name:
        gene = gene.split('*')[0]
    return gene


def get_parser():
    parser = argparse.ArgumentParser(description='Name an allele by comparing to a reference set. Construct a novel name containing SNPs if necessary. Use extract_refs to create suitable reference sets.')
    parser.add_argument('ref_set', help='reference set containing sequences from a single locus (e.g. IGH, TRB) (FASTA) (V genes should be IMGT-gapped)')
    parser.add_argument('sequence', help='input sequence')
    parser.add_argument('-g', '--gene', metavar='GENE_NAME', help='only consider reference sequences from the specified gene', dest='gene')
    parser.add_argument('-r', '--rev_comp', help='reverse-complement sequence before processing', action='store_true', dest='rev_comp')
    return parser


def main():
    args = get_parser().parse_args()

    gene_refs = simple.read_fasta(args.ref_set)

    locus = list(set([x[:3] for x in gene_refs.keys()]))

    if len(locus) == 0:
        print('No valid sequences found in the reference set. Please ensure that each sequence name starts with the three-letter locus name (e.g. IGH, TRB)')
        quit()
    elif len(locus) > 1:
        print('The reference set contains sequences from more than one locus. Please use a file that contains only sequences from a single locus (e.g. IGH, TRB)')
        quit()

    if args.gene:
        gene_refs = {k: v for k, v in gene_refs.items() if find_gene(k) == args.gene}
        
    if len(gene_refs) == 0:
        print(f"No reference genes to compare")
        quit()
        
    seq = args.sequence
    if args.rev_comp:
        seq = simple.reverse_complement(seq)

    novel_name, novel_seq, notes = novel_allele_name.name_novel(seq, gene_refs, len(seq) > 60)
           
    print(novel_name)
    
    if '.' in novel_seq:
        print(f"Gapped sequence:\n{novel_seq}")

    if notes:
        print(f"Notes:\n{notes}")



# Remove unwanted inserted codons in IMGT alignment of macaque IG sequences

from receptor_utils import simple_bio_seq as simple
import argparse

def get_parser():
    parser = argparse.ArgumentParser(description='Remove unwanted inserted codons in IMGT alignment of rhesus macaque IG sequences')
    parser.add_argument('infile', help='gapped file from IMGT')
    parser.add_argument('outfile', help='gapped file confirming to the customary alignment (without inserted codons)')
    parser.add_argument('chain', help='IGH, IGK or IGL')
    return parser


unwanted_codons = {
    'IGL': [21, 52, 53],
    'IGK': [21],
    'IGH': [16, 28],
}


def main():
    parser = get_parser()
    args = parser.parse_args()

    if args.chain not in ['IGH', 'IGK', 'IGL']:
        print("'chain' must be one of IGH, IGK, IGL")
        exit(0)

    seqs = simple.read_fasta(args.infile)
    unwanted_seqs = []

    for name, seq in seqs.items():
        for unw in sorted(unwanted_codons[args.chain], reverse=True):
            if seq[(unw-1)*3] == '.':
                seq = seq[:(unw-1)*3] + seq[3 + (unw-1)*3:]
            else:
                print(f"Unwanted codon at {unw} in sequence {name} is non-blank: dropping sequence")
                unwanted_seqs.append(name)
                break
            seqs[name] = seq

    for unw in unwanted_seqs:
        del seqs[unw]

    simple.write_fasta(args.outfile, seqs)


if __name__ == "__main__":
    main()
